@@ -27,7 +27,17 @@ from argparse import Namespace
 from collections.abc import ItemsView
 from functools import lru_cache
 from pathlib import Path
-from typing import Any, Callable, Dict, Generator, List, Optional, Tuple, Union
+from typing import (
+    Any,
+    Callable,
+    Dict,
+    Generator,
+    List,
+    Optional,
+    Sequence,
+    Tuple,
+    Union,
+)
 
 import yaml
 from ansible import constants
@@ -35,7 +45,7 @@ from ansible.errors import AnsibleError, AnsibleParserError
 from ansible.parsing.dataloader import DataLoader
 from ansible.parsing.mod_args import ModuleArgsParser
 from ansible.parsing.splitter import split_args
-from ansible.parsing.yaml.constructor import AnsibleConstructor
+from ansible.parsing.yaml.constructor import AnsibleConstructor, AnsibleMapping
 from ansible.parsing.yaml.loader import AnsibleLoader
 from ansible.parsing.yaml.objects import AnsibleBaseYAMLObject, AnsibleSequence
 from ansible.plugins.loader import add_all_plugin_dirs
@@ -63,7 +73,8 @@ from ansiblelint._internal.rules import (
 from ansiblelint.config import options
 from ansiblelint.constants import FileType
 from ansiblelint.errors import MatchError
-from ansiblelint.file_utils import Lintable, get_yaml_files
+from ansiblelint.file_utils import Lintable, discover_lintables
+from ansiblelint.text import removeprefix
 
 # ansible-lint doesn't need/want to know about encrypted secrets, so we pass a
 # string as the password to enable such yaml files to be opened and parsed
@@ -76,7 +87,7 @@ PLAYBOOK_DIR = os.environ.get('ANSIBLE_PLAYBOOK_DIR', None)
 _logger = logging.getLogger(__name__)
 
 
-def parse_yaml_from_file(filepath: str) -> Any:
+def parse_yaml_from_file(filepath: str) -> AnsibleBaseYAMLObject:
     dl = DataLoader()
     if hasattr(dl, 'set_vault_password'):
         dl.set_vault_password(DEFAULT_VAULT_PASSWORD)
@@ -89,7 +100,9 @@ def path_dwim(basedir: str, given: str) -> str:
     return str(dl.path_dwim(given))
 
 
-def ansible_template(basedir: str, varname: Any, templatevars, **kwargs) -> Any:
+def ansible_template(
+    basedir: str, varname: Any, templatevars: Any, **kwargs: Any
+) -> Any:
     dl = DataLoader()
     dl.set_basedir(basedir)
     templar = Templar(dl, variables=templatevars)
@@ -154,7 +167,7 @@ BLOCK_NAME_TO_ACTION_TYPE_MAP = {
 }
 
 
-def tokenize(line: str) -> Tuple[str, List[str], Dict]:
+def tokenize(line: str) -> Tuple[str, List[str], Dict[str, str]]:
     tokens = line.lstrip().split(" ")
     if tokens[0] == '-':
         tokens = tokens[1:]
@@ -175,18 +188,19 @@ def tokenize(line: str) -> Tuple[str, List[str], Dict]:
     return (command, args, kwargs)
 
 
-def _playbook_items(pb_data: dict) -> ItemsView:
+def _playbook_items(pb_data: AnsibleBaseYAMLObject) -> ItemsView:  # type: ignore
     if isinstance(pb_data, dict):
         return pb_data.items()
     if not pb_data:
-        return []
+        return []  # type: ignore
+
     # "if play" prevents failure if the play sequence contains None,
     # which is weird but currently allowed by Ansible
     # https://github.com/ansible-community/ansible-lint/issues/849
-    return [item for play in pb_data if play for item in play.items()]
+    return [item for play in pb_data if play for item in play.items()]  # type: ignore
 
 
-def _set_collections_basedir(basedir: str):
+def _set_collections_basedir(basedir: str) -> None:
     # Sets the playbook directory as playbook_paths for the collection loader
     try:
         # Ansible 2.10+
@@ -209,7 +223,7 @@ def find_children(lintable: Lintable) -> List[Lintable]:  # noqa: C901
     _set_collections_basedir(playbook_dir or os.path.abspath('.'))
     add_all_plugin_dirs(playbook_dir or '.')
     if lintable.kind == 'role':
-        playbook_ds = {'roles': [{'role': str(lintable.path)}]}
+        playbook_ds = AnsibleMapping({'roles': [{'role': str(lintable.path)}]})
     elif lintable.kind not in ("playbook", "tasks"):
         return []
     else:
@@ -221,8 +235,10 @@ def find_children(lintable: Lintable) -> List[Lintable]:  # noqa: C901
     basedir = os.path.dirname(str(lintable.path))
     # playbook_ds can be an AnsibleUnicode string, which we consider invalid
     if isinstance(playbook_ds, str):
-        raise MatchError(filename=str(lintable.path), rule=LoadingFailureRule)
+        raise MatchError(filename=str(lintable.path), rule=LoadingFailureRule())
     for item in _playbook_items(playbook_ds):
+        # if lintable.kind not in ["playbook"]:
+        #     continue
         for child in play_children(basedir, item, lintable.kind, playbook_dir):
             # We avoid processing parametrized children
             path_str = str(child.path)
@@ -246,7 +262,11 @@ def find_children(lintable: Lintable) -> List[Lintable]:  # noqa: C901
 
 
 def template(
-    basedir: str, value: Any, variables, fail_on_undefined=False, **kwargs
+    basedir: str,
+    value: Any,
+    variables: Any,
+    fail_on_undefined: bool = False,
+    **kwargs: str,
 ) -> Any:
     try:
         value = ansible_template(
@@ -264,7 +284,7 @@ def template(
 
 
 def play_children(
-    basedir: str, item: Tuple[str, Any], parent_type, playbook_dir
+    basedir: str, item: Tuple[str, Any], parent_type: FileType, playbook_dir: str
 ) -> List[Lintable]:
     delegate_map: Dict[str, Callable[[str, Any, Any, FileType], List[Lintable]]] = {
         'tasks': _taskshandlers_children,
@@ -294,7 +314,9 @@ def play_children(
     return []
 
 
-def _include_children(basedir: str, k, v, parent_type) -> List[Lintable]:
+def _include_children(
+    basedir: str, k: str, v: Any, parent_type: FileType
+) -> List[Lintable]:
     # handle special case include_tasks: name=filename.yml
     if k == 'include_tasks' and isinstance(v, dict) and 'file' in v:
         v = v['file']
@@ -308,7 +330,9 @@ def _include_children(basedir: str, k, v, parent_type) -> List[Lintable]:
     return [Lintable(result, kind=parent_type)]
 
 
-def _taskshandlers_children(basedir, k, v, parent_type: FileType) -> List[Lintable]:
+def _taskshandlers_children(
+    basedir: str, k: str, v: Union[None, Any], parent_type: FileType
+) -> List[Lintable]:
     results: List[Lintable] = []
     if v is None:
         raise MatchError(
@@ -364,9 +388,9 @@ def _taskshandlers_children(basedir, k, v, parent_type: FileType) -> List[Lintab
 
 
 def _get_task_handler_children_for_tasks_or_playbooks(
-    task_handler,
+    task_handler: Dict[str, Any],
     basedir: str,
-    k,
+    k: Any,
     parent_type: FileType,
 ) -> Lintable:
     """Try to get children of taskhandler for include/import tasks/playbooks."""
@@ -381,6 +405,7 @@ def _get_task_handler_children_for_tasks_or_playbooks(
             if not task_handler:
                 continue
 
+            # import pdb; pdb.set_trace()
             return Lintable(
                 path_dwim(basedir, task_handler[task_handler_key]), kind=child_type
             )
@@ -390,7 +415,7 @@ def _get_task_handler_children_for_tasks_or_playbooks(
     )
 
 
-def _validate_task_handler_action_for_role(th_action: dict) -> None:
+def _validate_task_handler_action_for_role(th_action: Dict[str, Any]) -> None:
     """Verify that the task handler action is valid for role include."""
     module = th_action['__ansible_module__']
 
@@ -404,7 +429,7 @@ def _validate_task_handler_action_for_role(th_action: dict) -> None:
 
 
 def _roles_children(
-    basedir: str, k, v, parent_type: FileType, main='main'
+    basedir: str, k: str, v: Sequence[Any], parent_type: FileType, main: str = 'main'
 ) -> List[Lintable]:
     results: List[Lintable] = []
     for role in v:
@@ -461,7 +486,9 @@ def _rolepath(basedir: str, role: str) -> Optional[str]:
     return role_path
 
 
-def _look_for_role_files(basedir: str, role: str, main='main') -> List[Lintable]:
+def _look_for_role_files(
+    basedir: str, role: str, main: Optional[str] = 'main'
+) -> List[Lintable]:
     role_path = _rolepath(basedir, role)
     if not role_path:
         return []
@@ -480,21 +507,12 @@ def _look_for_role_files(basedir: str, role: str, main='main') -> List[Lintable]
     return results
 
 
-def rolename(filepath):
-    idx = filepath.find('roles/')
-    if idx < 0:
-        return ''
-    role = filepath[idx + 6 :]
-    role = role[: role.find('/')]
-    return role
-
-
-def _kv_to_dict(v):
+def _kv_to_dict(v: str) -> Dict[str, Any]:
     (command, args, kwargs) = tokenize(v)
     return dict(__ansible_module__=command, __ansible_arguments__=args, **kwargs)
 
 
-def _sanitize_task(task: dict) -> dict:
+def _sanitize_task(task: Dict[str, Any]) -> Dict[str, Any]:
     """Return a stripped-off task structure compatible with new Ansible.
 
     This helper takes a copy of the incoming task and drops
@@ -510,7 +528,7 @@ def _sanitize_task(task: dict) -> dict:
 
 
 def normalize_task_v2(task: Dict[str, Any]) -> Dict[str, Any]:
-    """Ensure tasks have an action key and strings are converted to python objects."""
+    """Ensure tasks have a normalized action key and strings are converted to python objects."""
     result = dict()
 
     sanitized_task = _sanitize_task(task)
@@ -539,6 +557,13 @@ def normalize_task_v2(task: Dict[str, Any]) -> Dict[str, Any]:
             continue
         result[k] = v
 
+    if not isinstance(action, str):
+        raise RuntimeError("Task actions can only be strings, got %s" % action)
+    # convert builtin fqn calls to short forms because most rules know only
+    # about short calls but in the future we may switch the normalization to do
+    # the opposite. Mainly we currently consider normalized the module listing
+    # used by `ansible-doc -t module -l 2>/dev/null`
+    action = removeprefix(action, "ansible.builtin.")
     result['action'] = dict(__ansible_module__=action)
 
     if '_raw_params' in arguments:
@@ -555,54 +580,6 @@ def normalize_task_v2(task: Dict[str, Any]) -> Dict[str, Any]:
     return result
 
 
-def normalize_task_v1(task):  # noqa: C901
-    result = dict()
-    for (k, v) in task.items():
-        if k in VALID_KEYS or k.startswith('with_'):
-            if k in ('local_action', 'action'):
-                if not isinstance(v, dict):
-                    v = _kv_to_dict(v)
-                v['__ansible_arguments__'] = v.get('__ansible_arguments__', list())
-                result['action'] = v
-            else:
-                result[k] = v
-        else:
-            if isinstance(v, str):
-                v = _kv_to_dict(k + ' ' + v)
-            elif not v:
-                v = dict(__ansible_module__=k)
-            else:
-                if isinstance(v, dict):
-                    v.update(dict(__ansible_module__=k))
-                else:
-                    if k == '__line__':
-                        # Keep the line number stored
-                        result[k] = v
-                        continue
-                    # Tasks that include playbooks (rather than task files)
-                    # can get here
-                    # https://github.com/ansible-community/ansible-lint/issues/138
-                    raise RuntimeError(
-                        "Was not expecting value %s of type %s for key %s\n"
-                        "Task: %s. Check the syntax of your playbook using "
-                        "ansible-playbook --syntax-check"
-                        % (str(v), type(v), k, str(task))
-                    )
-            v['__ansible_arguments__'] = v.get('__ansible_arguments__', list())
-            result['action'] = v
-    if 'module' in result['action']:
-        # this happens when a task uses
-        # local_action:
-        #   module: ec2
-        #   etc...
-        result['action']['__ansible_module__'] = result['action']['module']
-        del result['action']['module']
-    if 'args' in result:
-        result['action'].update(result.get('args'))
-        del result['args']
-    return result
-
-
 def normalize_task(task: Dict[str, Any], filename: str) -> Dict[str, Any]:
     ansible_action_type = task.get('__ansible_action_type__', 'task')
     if '__ansible_action_type__' in task:
@@ -638,7 +615,9 @@ def task_to_str(task: Dict[str, Any]) -> str:
     return u"{0} {1}".format(action["__ansible_module__"], args)
 
 
-def extract_from_list(blocks, candidates: List[str]) -> List[Any]:
+def extract_from_list(
+    blocks: AnsibleBaseYAMLObject, candidates: List[str]
+) -> List[Any]:
     results = list()
     for block in blocks:
         for candidate in candidates:
@@ -653,7 +632,7 @@ def extract_from_list(blocks, candidates: List[str]) -> List[Any]:
     return results
 
 
-def add_action_type(actions, action_type: str) -> List[Any]:
+def add_action_type(actions: AnsibleBaseYAMLObject, action_type: str) -> List[Any]:
     results = list()
     for action in actions:
         # ignore empty task
@@ -664,7 +643,7 @@ def add_action_type(actions, action_type: str) -> List[Any]:
     return results
 
 
-def get_action_tasks(yaml, file: Lintable) -> List[Any]:
+def get_action_tasks(yaml: AnsibleBaseYAMLObject, file: Lintable) -> List[Any]:
     tasks = list()
     if file.kind in ['tasks', 'handlers']:
         tasks = add_action_type(yaml, file.kind)
@@ -690,7 +669,9 @@ def get_action_tasks(yaml, file: Lintable) -> List[Any]:
     ]
 
 
-def get_normalized_tasks(yaml, file: Lintable) -> List[Dict[str, Any]]:
+def get_normalized_tasks(
+    yaml: "AnsibleBaseYAMLObject", file: Lintable
+) -> List[Dict[str, Any]]:
     tasks = get_action_tasks(yaml, file)
     res = []
     for task in tasks:
@@ -712,14 +693,18 @@ def parse_yaml_linenumbers(lintable: Lintable) -> AnsibleBaseYAMLObject:
     The line numbers are stored in each node's LINE_NUMBER_KEY key.
     """
 
-    def compose_node(parent, index):
+    def compose_node(parent: yaml.nodes.Node, index: int) -> yaml.nodes.Node:
         # the line number where the previous token has ended (plus empty lines)
         line = loader.line
         node = Composer.compose_node(loader, parent, index)
-        node.__line__ = line + 1
+        if not isinstance(node, yaml.nodes.Node):
+            raise RuntimeError("Unexpected yaml data.")
+        setattr(node, '__line__', line + 1)
         return node
 
-    def construct_mapping(node, deep=False):
+    def construct_mapping(
+        node: AnsibleBaseYAMLObject, deep: bool = False
+    ) -> AnsibleMapping:
         mapping = AnsibleConstructor.construct_mapping(loader, node, deep=deep)
         if hasattr(node, '__line__'):
             mapping[LINE_NUMBER_KEY] = node.__line__
@@ -737,6 +722,7 @@ def parse_yaml_linenumbers(lintable: Lintable) -> AnsibleBaseYAMLObject:
         loader.construct_mapping = construct_mapping
         data = loader.get_single_data()
     except (yaml.parser.ParserError, yaml.scanner.ScannerError) as e:
+        logging.exception(e)
         raise SystemExit("Failed to parse YAML in %s: %s" % (lintable.path, str(e)))
     return data
 
@@ -752,6 +738,17 @@ def get_first_cmd_arg(task: Dict[str, Any]) -> Any:
     return first_cmd_arg
 
 
+def get_second_cmd_arg(task: Dict[str, Any]) -> Any:
+    try:
+        if 'cmd' in task['action']:
+            second_cmd_arg = task['action']['cmd'].split()[1]
+        else:
+            second_cmd_arg = task['action']['__ansible_arguments__'][1]
+    except IndexError:
+        return None
+    return second_cmd_arg
+
+
 def is_playbook(filename: str) -> bool:
     """
     Check if the file is a playbook.
@@ -813,7 +810,7 @@ def get_lintables(
             lintables.append(lintable)
     else:
 
-        for filename in get_yaml_files(options):
+        for filename in discover_lintables(options):
 
             p = Path(filename)
             # skip exclusions
@@ -844,7 +841,7 @@ def _extend_with_roles(lintables: List[Lintable]) -> None:
             role = lintable.path
             while role.parent.name != "roles" and role.name:
                 role = role.parent
-            if role.exists:
+            if role.exists and not role.is_file():
                 lintable = Lintable(role, kind="role")
                 if lintable not in lintables:
                     _logger.debug("Added role: %s", lintable)
